@@ -20,9 +20,11 @@ import (
 	"github.com/containers/image/v5/types"
 	"github.com/containers/podman/v3/libpod"
 	"github.com/containers/podman/v3/pkg/api/handlers/utils"
+	api "github.com/containers/podman/v3/pkg/api/types"
 	"github.com/containers/podman/v3/pkg/auth"
 	"github.com/containers/podman/v3/pkg/channel"
 	"github.com/containers/storage/pkg/archive"
+	"github.com/docker/docker/pkg/jsonmessage"
 	"github.com/gorilla/schema"
 	"github.com/opencontainers/runtime-spec/specs-go"
 	"github.com/pkg/errors"
@@ -34,13 +36,16 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 		contentType := hdr[0]
 		switch contentType {
 		case "application/tar":
-			logrus.Warnf("tar file content type is  %s, should use \"application/x-tar\" content type", contentType)
+			logrus.Infof("tar file content type is  %s, should use \"application/x-tar\" content type", contentType)
 		case "application/x-tar":
 			break
 		default:
-			utils.BadRequest(w, "Content-Type", hdr[0],
-				fmt.Errorf("Content-Type: %s is not supported. Should be \"application/x-tar\"", hdr[0]))
-			return
+			if utils.IsLibpodRequest(r) {
+				utils.BadRequest(w, "Content-Type", hdr[0],
+					fmt.Errorf("Content-Type: %s is not supported. Should be \"application/x-tar\"", hdr[0]))
+				return
+			}
+			logrus.Infof("tar file content type is  %s, should use \"application/x-tar\" content type", contentType)
 		}
 	}
 
@@ -73,10 +78,12 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 		CacheFrom              string   `schema:"cachefrom"`
 		Compression            uint64   `schema:"compression"`
 		ConfigureNetwork       string   `schema:"networkmode"`
-		CpuPeriod              uint64   `schema:"cpuperiod"`  // nolint
-		CpuQuota               int64    `schema:"cpuquota"`   // nolint
-		CpuSetCpus             string   `schema:"cpusetcpus"` // nolint
-		CpuShares              uint64   `schema:"cpushares"`  // nolint
+		CpuPeriod              uint64   `schema:"cpuperiod"`    // nolint
+		CpuQuota               int64    `schema:"cpuquota"`     // nolint
+		CpuSetCpus             string   `schema:"cpusetcpus"`   // nolint
+		CpuSetMems             string   `schema:"cpusetmems"`   // nolint
+		CpuShares              uint64   `schema:"cpushares"`    // nolint
+		CgroupParent           string   `schema:"cgroupparent"` // nolint
 		DNSOptions             string   `schema:"dnsoptions"`
 		DNSSearch              string   `schema:"dnssearch"`
 		DNSServers             string   `schema:"dnsservers"`
@@ -100,12 +107,13 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 		NamespaceOptions       string   `schema:"nsoptions"`
 		NoCache                bool     `schema:"nocache"`
 		OutputFormat           string   `schema:"outputformat"`
-		Platform               string   `schema:"platform"`
+		Platform               []string `schema:"platform"`
 		Pull                   bool     `schema:"pull"`
 		PullPolicy             string   `schema:"pullpolicy"`
 		Quiet                  bool     `schema:"q"`
 		Registry               string   `schema:"registry"`
 		Rm                     bool     `schema:"rm"`
+		RusageLogFile          string   `schema:"rusagelogfile"`
 		Seccomp                string   `schema:"seccomp"`
 		SecurityOpt            string   `schema:"securityopt"`
 		ShmSize                int      `schema:"shmsize"`
@@ -119,15 +127,23 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 		Registry:   "docker.io",
 		Rm:         true,
 		ShmSize:    64 * 1024 * 1024,
-		Tag:        []string{},
 	}
 
-	decoder := r.Context().Value("decoder").(*schema.Decoder)
+	decoder := r.Context().Value(api.DecoderKey).(*schema.Decoder)
 	if err := decoder.Decode(&query, r.URL.Query()); err != nil {
 		utils.Error(w, http.StatusText(http.StatusBadRequest), http.StatusBadRequest, err)
 		return
 	}
 
+	// if layers field not set assume its not from a valid podman-client
+	// could be a docker client, set `layers=true` since that is the default
+	// expected behviour
+	if !utils.IsLibpodRequest(r) {
+		if _, found := r.URL.Query()["layers"]; !found {
+			query.Layers = true
+		}
+	}
+
 	// convert addcaps formats
 	var addCaps = []string{}
 	if _, found := r.URL.Query()["addcaps"]; found {
@@ -145,22 +161,19 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 		var m = []string{}
 		if err := json.Unmarshal([]byte(query.Dockerfile), &m); err != nil {
 			// it's not json, assume just a string
-			m = append(m, query.Dockerfile)
+			m = []string{filepath.Join(contextDirectory, query.Dockerfile)}
 		}
 		containerFiles = m
 	} else {
-		containerFiles = []string{"Dockerfile"}
+		containerFiles = []string{filepath.Join(contextDirectory, "Dockerfile")}
 		if utils.IsLibpodRequest(r) {
-			containerFiles = []string{"Containerfile"}
-			if _, err = os.Stat(filepath.Join(contextDirectory, "Containerfile")); err != nil {
-				if _, err1 := os.Stat(filepath.Join(contextDirectory, "Dockerfile")); err1 == nil {
-					containerFiles = []string{"Dockerfile"}
-				} else {
+			containerFiles = []string{filepath.Join(contextDirectory, "Containerfile")}
+			if _, err = os.Stat(containerFiles[0]); err != nil {
+				containerFiles = []string{filepath.Join(contextDirectory, "Dockerfile")}
+				if _, err1 := os.Stat(containerFiles[0]); err1 != nil {
 					utils.BadRequest(w, "dockerfile", query.Dockerfile, err)
 				}
 			}
-		} else {
-			containerFiles = []string{"Dockerfile"}
 		}
 	}
 
@@ -189,8 +202,8 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 	var devices = []string{}
 	if _, found := r.URL.Query()["devices"]; found {
 		var m = []string{}
-		if err := json.Unmarshal([]byte(query.DropCapabilities), &m); err != nil {
-			utils.BadRequest(w, "devices", query.DropCapabilities, err)
+		if err := json.Unmarshal([]byte(query.Devices), &m); err != nil {
+			utils.BadRequest(w, "devices", query.Devices, err)
 			return
 		}
 		devices = m
@@ -392,19 +405,19 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 	defer auth.RemoveAuthfile(authfile)
 
 	// Channels all mux'ed in select{} below to follow API build protocol
-	stdout := channel.NewWriter(make(chan []byte, 1))
+	stdout := channel.NewWriter(make(chan []byte))
 	defer stdout.Close()
 
-	auxout := channel.NewWriter(make(chan []byte, 1))
+	auxout := channel.NewWriter(make(chan []byte))
 	defer auxout.Close()
 
-	stderr := channel.NewWriter(make(chan []byte, 1))
+	stderr := channel.NewWriter(make(chan []byte))
 	defer stderr.Close()
 
-	reporter := channel.NewWriter(make(chan []byte, 1))
+	reporter := channel.NewWriter(make(chan []byte))
 	defer reporter.Close()
 
-	runtime := r.Context().Value("runtime").(*libpod.Runtime)
+	runtime := r.Context().Value(api.RuntimeKey).(*libpod.Runtime)
 	rtc, err := runtime.GetConfig()
 	if err != nil {
 		utils.Error(w, "Something went wrong.", http.StatusInternalServerError, errors.Wrap(err, "Decode()"))
@@ -421,7 +434,9 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 			CPUPeriod:          query.CpuPeriod,
 			CPUQuota:           query.CpuQuota,
 			CPUSetCPUs:         query.CpuSetCpus,
+			CPUSetMems:         query.CpuSetMems,
 			CPUShares:          query.CpuShares,
+			CgroupParent:       query.CgroupParent,
 			DNSOptions:         dnsoptions,
 			DNSSearch:          dnssearch,
 			DNSServers:         dnsservers,
@@ -463,6 +478,7 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 		Registry:                       registry,
 		RemoveIntermediateCtrs:         query.Rm,
 		ReportWriter:                   reporter,
+		RusageLogFile:                  query.RusageLogFile,
 		Squash:                         query.Squash,
 		Target:                         query.Target,
 		SystemContext: &types.SystemContext{
@@ -471,16 +487,17 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 		},
 	}
 
-	if len(query.Platform) > 0 {
-		variant := ""
-		buildOptions.OS, buildOptions.Architecture, variant, err = parse.Platform(query.Platform)
+	for _, platformSpec := range query.Platform {
+		os, arch, variant, err := parse.Platform(platformSpec)
 		if err != nil {
-			utils.BadRequest(w, "platform", query.Platform, err)
+			utils.BadRequest(w, "platform", platformSpec, err)
 			return
 		}
-		buildOptions.SystemContext.OSChoice = buildOptions.OS
-		buildOptions.SystemContext.ArchitectureChoice = buildOptions.Architecture
-		buildOptions.SystemContext.VariantChoice = variant
+		buildOptions.Platforms = append(buildOptions.Platforms, struct{ OS, Arch, Variant string }{
+			OS:      os,
+			Arch:    arch,
+			Variant: variant,
+		})
 	}
 	if _, found := r.URL.Query()["timestamp"]; found {
 		ts := time.Unix(query.Timestamp, 0)
@@ -511,7 +528,7 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 
 	// Send headers and prime client for stream to come
 	w.WriteHeader(http.StatusOK)
-	w.Header().Add("Content-Type", "application/json")
+	w.Header().Set("Content-Type", "application/json")
 	flush()
 
 	body := w.(io.Writer)
@@ -527,11 +544,13 @@ func BuildImage(w http.ResponseWriter, r *http.Request) {
 
 	enc := json.NewEncoder(body)
 	enc.SetEscapeHTML(true)
-loop:
+
 	for {
 		m := struct {
-			Stream string `json:"stream,omitempty"`
-			Error  string `json:"error,omitempty"`
+			Stream string                 `json:"stream,omitempty"`
+			Error  *jsonmessage.JSONError `json:"errorDetail,omitempty"`
+			// NOTE: `error` is being deprecated check https://github.com/moby/moby/blob/master/pkg/jsonmessage/jsonmessage.go#L148
+			ErrorMessage string `json:"error,omitempty"` // deprecate this slowly
 		}{}
 
 		select {
@@ -541,26 +560,29 @@ loop:
 				stderr.Write([]byte(err.Error()))
 			}
 			flush()
-		case e := <-auxout.Chan():
-			m.Stream = string(e)
-			if err := enc.Encode(m); err != nil {
-				stderr.Write([]byte(err.Error()))
-			}
-			flush()
 		case e := <-reporter.Chan():
 			m.Stream = string(e)
 			if err := enc.Encode(m); err != nil {
 				stderr.Write([]byte(err.Error()))
 			}
 			flush()
+		case e := <-auxout.Chan():
+			m.Stream = string(e)
+			if err := enc.Encode(m); err != nil {
+				stderr.Write([]byte(err.Error()))
+			}
+			flush()
 		case e := <-stderr.Chan():
-			m.Error = string(e)
+			m.ErrorMessage = string(e)
+			m.Error = &jsonmessage.JSONError{
+				Message: m.ErrorMessage,
+			}
 			if err := enc.Encode(m); err != nil {
 				logrus.Warnf("Failed to json encode error %v", err)
 			}
 			flush()
+			return
 		case <-runCtx.Done():
-			flush()
 			if success {
 				if !utils.IsLibpodRequest(r) {
 					m.Stream = fmt.Sprintf("Successfully built %12.12s\n", imageID)
@@ -577,7 +599,8 @@ loop:
 					}
 				}
 			}
-			break loop
+			flush()
+			return
 		case <-r.Context().Done():
 			cancel()
 			logrus.Infof("Client disconnect reported for build %q / %q.", registry, query.Dockerfile)
